from __future__ import annotations

import os
import torch

import pytest

from torch_ivf.index import IndexFlatL2, IndexIVFFlat, SearchParams


def _toy_data(d=16, nb=512, nq=8, seed=0):
    g = torch.Generator().manual_seed(seed)
    xb = torch.randn(nb, d, generator=g)
    xq = torch.randn(nq, d, generator=g)
    return xb, xq


def test_ivf_train_add_and_search_matches_flat_when_nprobe_full():
    d = 16
    xb, xq = _toy_data(d=d, nb=400, nq=6)
    index = IndexIVFFlat(d, nlist=16)
    index.train(xb)
    index.add(xb)
    index.nprobe = index.nlist

    D_ivf, I_ivf = index.search(xq, k=5)

    flat = IndexFlatL2(d)
    flat.add(xb)
    D_flat, I_flat = flat.search(xq, k=5)

    assert torch.allclose(D_ivf.cpu(), D_flat.cpu(), atol=1e-5)
    assert torch.equal(I_ivf.cpu(), I_flat.cpu())


def test_ivf_search_csr_matches_flat_when_nprobe_full():
    d = 16
    xb, xq = _toy_data(d=d, nb=400, nq=6, seed=2)
    index = IndexIVFFlat(d, nlist=16)
    index.train(xb)
    index.add(xb)
    index.nprobe = index.nlist
    index.search_mode = "csr"

    D_ivf, I_ivf = index.search(xq, k=5)

    flat = IndexFlatL2(d)
    flat.add(xb)
    D_flat, I_flat = flat.search(xq, k=5)

    assert torch.allclose(D_ivf.cpu(), D_flat.cpu(), atol=1e-5)
    assert torch.equal(I_ivf.cpu(), I_flat.cpu())


def test_ivf_search_auto_matches_flat_when_nprobe_full():
    d = 16
    xb, xq = _toy_data(d=d, nb=400, nq=6, seed=3)
    index = IndexIVFFlat(d, nlist=16)
    index.train(xb)
    index.add(xb)
    index.nprobe = index.nlist
    index.search_mode = "auto"

    D_ivf, I_ivf = index.search(xq, k=5)

    flat = IndexFlatL2(d)
    flat.add(xb)
    D_flat, I_flat = flat.search(xq, k=5)

    assert torch.allclose(D_ivf.cpu(), D_flat.cpu(), atol=1e-5)
    assert torch.equal(I_ivf.cpu(), I_flat.cpu())


def test_ivf_range_search_respects_radius():
    xb, xq = _toy_data(d=8, nb=200, nq=3, seed=1)
    index = IndexIVFFlat(8, nlist=8)
    index.train(xb)
    index.add(xb)
    lims, dvals, ids = index.range_search(xq, radius=5.0)
    assert lims.shape[0] == xq.shape[0] + 1
    assert dvals.shape[0] == ids.shape[0]
    assert torch.all(lims[1:] >= lims[:-1])
    assert lims[-1] > 0


def test_ivf_state_dict_roundtrip(tmp_path):
    xb, _ = _toy_data(nb=64)
    index = IndexIVFFlat(16, nlist=4)
    index.train(xb)
    index.add(xb)
    path = tmp_path / "ivf.pt"
    index.save(path.as_posix())

    loaded = IndexIVFFlat.load(path.as_posix())
    assert loaded.nlist == index.nlist
    assert loaded.nprobe == index.nprobe
    assert loaded.ntotal == index.ntotal


def test_ivf_add_without_train_raises():
    index = IndexIVFFlat(16, nlist=4)
    with pytest.raises(RuntimeError):
        index.add(torch.randn(4, 16))


def test_ivf_max_codes_prefix_boundary_cases():
    index = IndexIVFFlat(1, nlist=4)
    offsets = torch.tensor([0, 3, 7, 12, 12], dtype=torch.long, device=index.device)
    index._list_offsets = offsets
    index._list_offsets_cpu = offsets.to("cpu").tolist()

    top_lists = torch.tensor([[0, 1, 2]], dtype=torch.long, device=index.device)

    for max_codes, expected_lists, expected_probes in [
        (6, [0], [0]),
        (7, [0, 1], [0, 1]),
        (8, [0, 1], [0, 1]),
    ]:
        index.max_codes = max_codes

        tasks_q, tasks_l, _ = index._build_tasks_from_lists(top_lists)
        assert tasks_q.to("cpu").tolist() == [0] * len(expected_lists)
        assert tasks_l.to("cpu").tolist() == expected_lists

        tasks_q, tasks_l, tasks_p, _ = index._build_tasks_from_lists_with_probe(top_lists)
        assert tasks_q.to("cpu").tolist() == [0] * len(expected_lists)
        assert tasks_l.to("cpu").tolist() == expected_lists
        assert tasks_p.to("cpu").tolist() == expected_probes


def test_ivf_search_csr_matches_matrix_when_max_codes_unlimited():
    d = 16
    xb, xq = _toy_data(d=d, nb=400, nq=16, seed=4)
    index = IndexIVFFlat(d, nlist=16, nprobe=4)
    index.train(xb)
    index.add(xb)
    index.max_codes = 0

    index.search_mode = "matrix"
    D_matrix, I_matrix = index.search(xq, k=5)

    index.search_mode = "csr"
    D_csr, I_csr = index.search(xq, k=5)

    assert torch.allclose(D_matrix.cpu(), D_csr.cpu(), atol=1e-5)
    assert torch.equal(I_matrix.cpu(), I_csr.cpu())


def test_ivf_search_cache_invalidated_on_add_and_max_codes():
    xb, xq = _toy_data(d=8, nb=64, nq=4, seed=5)
    index = IndexIVFFlat(8, nlist=8)
    index.train(xb)
    index.add(xb)
    index.search_mode = "csr"

    index.max_codes = 10
    index.search(xq, k=3)
    assert index._effective_max_codes_cache is not None

    index.max_codes = 5
    assert index._effective_max_codes_cache is None

    index.search(xq, k=3)
    assert index._list_sizes is not None

    index.add(xb[:8])
    assert index._list_sizes is None
    assert index._effective_max_codes_cache is None


def test_ivf_to_invalidates_search_cache():
    xb, xq = _toy_data(d=8, nb=64, nq=4, seed=11)
    index = IndexIVFFlat(8, nlist=8)
    index.train(xb)
    index.add(xb)
    index.search_mode = "csr"
    index.max_codes = 10
    index.search(xq, k=3)

    assert index._centroids_t is not None
    assert index._centroid_norm2 is not None
    assert index._list_sizes is not None
    assert index._list_sizes_cpu is not None
    assert index._effective_max_codes_cache is not None

    moved = index.to(index.device)
    assert moved._centroids_t is None
    assert moved._centroid_norm2 is None
    assert moved._list_sizes is None
    assert moved._list_sizes_cpu is None
    assert moved._effective_max_codes_cache is None


def test_ivf_workspace_reuse_does_not_change_results():
    xb, xq = _toy_data(d=16, nb=128, nq=8, seed=6)
    index = IndexIVFFlat(16, nlist=32, nprobe=8)
    index.train(xb)
    index.add(xb)
    index.search_mode = "csr"

    d1, i1 = index.search(xq, k=5)
    caps = dict(index._workspace.capacity)
    d2, i2 = index.search(xq, k=5)

    assert torch.allclose(d1.cpu(), d2.cpu(), atol=1e-6)
    assert torch.equal(i1.cpu(), i2.cpu())
    for name, cap in caps.items():
        assert index._workspace.capacity[name] >= cap


def test_search_params_resolve_clamps_nprobe_and_prefers_explicit():
    index = IndexIVFFlat(2, nlist=4, nprobe=2)
    params = SearchParams(
        profile="approx",
        approximate=True,
        nprobe=10,
        max_codes=7,
        candidate_budget=16,
        budget_strategy="uniform",
    )
    config = index._resolve_search_params(params)
    assert config.nprobe == 4
    assert config.max_codes == 7
    assert config.candidate_budget == 16


def test_search_params_profile_presets_apply_per_list_budget():
    index = IndexIVFFlat(2, nlist=4, nprobe=2)
    params = SearchParams(profile="approx_quality")
    config = index._resolve_search_params(params)
    assert config.approximate is True
    assert config.use_per_list_sizes is True
    assert config.candidate_budget == 131072
    assert config.list_ordering == "residual_norm_asc"


def test_candidate_budget_dynamic_nprobe_skips_far_lists():
    index = IndexIVFFlat(2, nlist=4)
    offsets = torch.tensor([0, 10, 20, 30, 40], dtype=torch.long, device=index.device)
    index._list_offsets = offsets
    index._list_sizes = None

    top_lists = torch.tensor([[0, 1, 2, 3]], dtype=torch.long, device=index.device)
    top_scores = torch.tensor([[0.1, 1.0, 2.0, 3.0]], dtype=torch.float32, device=index.device)

    params = SearchParams(
        profile="approx",
        approximate=True,
        nprobe=4,
        candidate_budget=2,
        budget_strategy="distance_weighted",
        dynamic_nprobe=True,
        min_codes_per_list=1,
        use_per_list_sizes=True,
    )
    config = index._resolve_search_params(params)
    sizes = index._allocate_candidate_sizes(top_lists, top_scores, config)
    sizes_cpu = sizes.to("cpu").tolist()[0]
    assert sizes_cpu[0] > 0
    assert sizes_cpu[1] > 0
    assert sizes_cpu[2] == 0
    assert sizes_cpu[3] == 0
    assert sum(sizes_cpu) <= 2


def test_rebuild_lists_residual_norm_orders_within_list():
    index = IndexIVFFlat(1, nlist=2)
    index._centroids = torch.tensor([[0.0], [10.0]], dtype=torch.float32, device=index.device)
    index._packed_embeddings = torch.tensor([[3.0], [1.0], [2.0], [10.0], [12.0]], device=index.device)
    index._packed_norms = (index._packed_embeddings * index._packed_embeddings).sum(dim=1)
    index._list_ids = torch.tensor([30, 10, 20, 100, 120], dtype=torch.long, device=index.device)
    index._list_offsets = torch.tensor([0, 3, 5], dtype=torch.long, device=index.device)
    index._list_offsets_cpu = [0, 3, 5]
    index._ntotal = 5
    index._is_trained = True

    index.rebuild_lists(ordering="residual_norm_asc")

    list0 = index._packed_embeddings[:3].to("cpu").flatten().tolist()
    list0_ids = index._list_ids[:3].to("cpu").tolist()
    assert list0 == [1.0, 2.0, 3.0]
    assert list0_ids == [10, 20, 30]


def test_ivf_csr_blocked_groups_matches_unblocked():
    if os.environ.get("TORCH_IVF_RUN_GPU_EXPERIMENTAL_TESTS") != "1":
        pytest.skip("Set TORCH_IVF_RUN_GPU_EXPERIMENTAL_TESTS=1 to run experimental GPU tests.")

    if not torch.cuda.is_available():
        pytest.skip("CUDA/ROCm device is required for blocked CSR test.")
    d = 32
    xb, xq = _toy_data(d=d, nb=1024, nq=512, seed=7)
    device = torch.device("cuda")
    xb = xb.to(device)
    xq = xq.to(device)

    index = IndexIVFFlat(d, nlist=32, nprobe=8, device=device)
    index.train(xb)
    index.add(xb)
    index.search_mode = "csr"
    index._csr_small_batch_avg_group_threshold = 0.0

    params = SearchParams(
        profile="approx",
        approximate=True,
        candidate_budget=8192,
        budget_strategy="uniform",
        use_per_list_sizes=True,
        debug_stats=True,
    )

    def run(block_size: int):
        index._csr_list_block_size = lambda: block_size
        index._csr_block_pad_ratio_limit = lambda: 10.0
        index._csr_block_cost_ratio_limit = lambda: 100.0
        index._csr_block_max_elements = lambda: 1_000_000_000
        dists, ids = index.search(xq, k=5, params=params)
        stats = index.last_search_stats
        return dists, ids, stats

    d_block, i_block, stats_block = run(4)
    assert stats_block is not None
    assert stats_block.get("blocked_blocks", 0) > 0

    d_list, i_list, _ = run(1)
    assert torch.allclose(d_block.cpu(), d_list.cpu(), atol=1e-5)
    assert torch.equal(i_block.cpu(), i_list.cpu())


def test_ivf_csr_buffered_blocked_matches_unblocked_exact():
    # This test exercises an optional, experimental GPU batching path.
    # On some ROCm/Windows setups, pytest can hang intermittently around GPU teardown.
    # Gate it behind an explicit opt-in env var so the default suite stays reliable.
    if os.environ.get("TORCH_IVF_RUN_GPU_EXPERIMENTAL_TESTS") != "1":
        pytest.skip("Set TORCH_IVF_RUN_GPU_EXPERIMENTAL_TESTS=1 to run experimental GPU tests.")

    if not torch.cuda.is_available():
        pytest.skip("CUDA/ROCm device is required for buffered-blocked CSR test.")

    d = 32
    xb, xq = _toy_data(d=d, nb=4096, nq=1024, seed=11)
    device = torch.device("cuda")
    xb = xb.to(device)
    xq = xq.to(device)

    index = IndexIVFFlat(d, nlist=64, nprobe=16, device=device)
    index.train(xb[:2048], generator=torch.Generator(device="cpu").manual_seed(123))
    index.add(xb)
    index.search_mode = "csr"
    index._csr_small_batch_avg_group_threshold = 0.0

    # Force the blocked-buffered path to be considered.
    index._csr_buf_blocked_enabled = True
    index._csr_list_block_size = lambda: 16
    index._csr_block_pad_ratio_limit = lambda: 10.0
    index._csr_block_cost_ratio_limit = lambda: 100.0
    index._csr_block_max_elements = lambda: 1_000_000_000

    index._csr_buf_blocked_enabled = False
    d_ref, i_ref = index.search(xq, k=10)

    index._csr_buf_blocked_enabled = True
    d_blk, i_blk = index.search(xq, k=10)

    assert torch.allclose(d_ref.cpu(), d_blk.cpu(), atol=1e-5)
    assert torch.equal(i_ref.cpu(), i_blk.cpu())



import importlib


def test_faiss_cpu_available():
    module = importlib.import_module("faiss")
    assert hasattr(module, "IndexFlatL2")


def test_pytest_runs():
    # Dummy assertion to ensure pytest discovers at least one test case.
    assert True



from __future__ import annotations

import numpy as np
import torch

import faiss
import pytest

from torch_ivf.index import IndexFlatIP, IndexFlatL2


def _generate_data(d=16, nb=256, nq=4, *, seed=0):
    g = torch.Generator().manual_seed(seed)
    xb = torch.randn(nb, d, generator=g)
    xq = torch.randn(nq, d, generator=g)
    return xb, xq


def _as_numpy(x: torch.Tensor) -> np.ndarray:
    return x.cpu().numpy().astype("float32")


@pytest.mark.parametrize("k", [1, 5])
def test_index_flat_l2_matches_faiss(k):
    d = 16
    xb, xq = _generate_data(d=d, nb=128, nq=8)

    torch_index = IndexFlatL2(d)
    torch_index.add(xb)
    Dt, It = torch_index.search(xq, k)

    faiss_index = faiss.IndexFlatL2(d)
    faiss_index.add(_as_numpy(xb))
    Df, If = faiss_index.search(_as_numpy(xq), k)

    assert torch.allclose(Dt.cpu(), torch.from_numpy(Df), atol=1e-5)
    assert torch.equal(It.cpu(), torch.from_numpy(If))


def test_index_flat_ip_matches_faiss():
    d = 32
    xb, xq = _generate_data(d=d, nb=200, nq=5)

    torch_index = IndexFlatIP(d)
    torch_index.add(xb)
    Dt, It = torch_index.search(xq, k=10)

    faiss_index = faiss.IndexFlatIP(d)
    faiss_index.add(_as_numpy(xb))
    Df, If = faiss_index.search(_as_numpy(xq), 10)

    assert torch.allclose(Dt.cpu(), torch.from_numpy(Df), atol=1e-5)
    assert torch.equal(It.cpu(), torch.from_numpy(If))


def test_add_with_ids_and_range_search_l2():
    d = 8
    xb, xq = _generate_data(d=d, nb=50, nq=2)
    custom_ids = torch.arange(100, 150, dtype=torch.long)

    torch_index = IndexFlatL2(d)
    torch_index.add_with_ids(xb, custom_ids)

    lims, Dvals, Lvals = torch_index.range_search(xq, radius=5.0)
    faiss_index = faiss.IndexFlatL2(d)
    faiss_index.add(_as_numpy(xb))
    lims_ref, D_ref, L_ref = faiss_index.range_search(_as_numpy(xq), 5.0)

    lims_ref_t = torch.from_numpy(lims_ref.astype("int64"))
    assert torch.equal(lims.cpu(), lims_ref_t)
    assert torch.allclose(Dvals.cpu(), torch.from_numpy(D_ref), atol=1e-5)
    # Faiss returns base indices; convert to provided IDs.
    base_idx = torch.from_numpy(L_ref.astype("int64"))
    faiss_ids = custom_ids[base_idx]
    assert torch.equal(Lvals.cpu(), faiss_ids)


def test_reset_clears_state():
    idx = IndexFlatIP(4)
    idx.add(torch.randn(10, 4))
    assert idx.ntotal == 10
    idx.reset()
    assert idx.ntotal == 0
    D, I = idx.search(torch.randn(2, 4), k=3)
    assert torch.isinf(D).all() or torch.isneginf(D).all()
    assert (I == -1).all()


def test_index_flat_chunked_l2_matches_faiss():
    d = 16
    xb, xq = _generate_data(d=d, nb=500, nq=7, seed=4)
    torch_index = IndexFlatL2(d)
    torch_index.search_mode = "chunked"
    torch_index.add(xb)
    Dt, It = torch_index.search(xq, k=5)

    faiss_index = faiss.IndexFlatL2(d)
    faiss_index.add(_as_numpy(xb))
    Df, If = faiss_index.search(_as_numpy(xq), 5)

    assert torch.allclose(Dt.cpu(), torch.from_numpy(Df), atol=1e-5)
    assert torch.equal(It.cpu(), torch.from_numpy(If))


def test_index_flat_auto_l2_matches_faiss_when_forced_chunked():
    d = 16
    xb, xq = _generate_data(d=d, nb=500, nq=7, seed=5)
    torch_index = IndexFlatL2(d)
    torch_index.search_mode = "auto"
    torch_index.auto_matrix_bytes_threshold = 1
    torch_index.add(xb)
    Dt, It = torch_index.search(xq, k=5)

    faiss_index = faiss.IndexFlatL2(d)
    faiss_index.add(_as_numpy(xb))
    Df, If = faiss_index.search(_as_numpy(xq), 5)

    assert torch.allclose(Dt.cpu(), torch.from_numpy(Df), atol=1e-5)
    assert torch.equal(It.cpu(), torch.from_numpy(If))


def test_index_flat_chunked_ip_matches_faiss():
    d = 32
    xb, xq = _generate_data(d=d, nb=600, nq=6, seed=6)
    torch_index = IndexFlatIP(d)
    torch_index.search_mode = "chunked"
    torch_index.add(xb)
    Dt, It = torch_index.search(xq, k=10)

    faiss_index = faiss.IndexFlatIP(d)
    faiss_index.add(_as_numpy(xb))
    Df, If = faiss_index.search(_as_numpy(xq), 10)

    assert torch.allclose(Dt.cpu(), torch.from_numpy(Df), atol=1e-5)
    assert torch.equal(It.cpu(), torch.from_numpy(If))



"""torch_ivf package initialization."""

try:
    import torch as _torch  # noqa: F401
except Exception as e:  # pragma: no cover
    raise ImportError(
        "torch-ivf requires a working PyTorch installation, but importing 'torch' failed. "
        "Install PyTorch first (choose CPU/CUDA/ROCm/DirectML as appropriate), then install torch-ivf. "
        "If you want a simple CPU setup, you can also install torch-ivf with the 'pytorch' extra: "
        "pip install \"torch-ivf[pytorch]\". "
        f"Original error: {e!r}"
    ) from e

__all__ = [
    "index",
    "nn",
    "utils",
]



from __future__ import annotations

import torch

from .base import IndexBase, MetricType


class IndexFlat(IndexBase):
    """Base implementation shared by IndexFlatL2/IP."""

    def __init__(
        self,
        d: int,
        *,
        metric: MetricType = "l2",
        device: torch.device | str | None = None,
        dtype: torch.dtype = torch.float32,
    ) -> None:
        super().__init__(d, metric=metric, device=device, dtype=dtype)
        self._xb = torch.empty((0, d), dtype=self.dtype, device=self.device)
        self._xb_norms = torch.empty(0, dtype=self.dtype, device=self.device)
        self._ids = torch.empty(0, dtype=torch.long, device=self.device)
        self._is_trained = True  # Flat indexes do not require training.
        self._search_mode = "matrix"
        self._auto_matrix_bytes_threshold = 256 * 1024 * 1024

    # ------------------------------------------------------------------ #
    # IndexBase requirements
    # ------------------------------------------------------------------ #
    def train(self, xb: torch.Tensor) -> None:
        self._validate_input(xb)
        self._is_trained = True

    def add(self, xb: torch.Tensor) -> None:
        xb = self._validate_input(xb)
        if xb.shape[0] == 0:
            return
        new_ids = torch.arange(
            self._ntotal,
            self._ntotal + xb.shape[0],
            dtype=torch.long,
            device=self.device,
        )
        self._append(xb, new_ids)

    def add_with_ids(self, xb: torch.Tensor, ids: torch.Tensor) -> None:
        xb = self._validate_input(xb)
        ids = self._validate_ids(ids, xb.shape[0])
        self._append(xb, ids.to(self.device))

    @property
    def search_mode(self) -> str:
        return self._search_mode

    @search_mode.setter
    def search_mode(self, value: str) -> None:
        if value not in {"matrix", "chunked", "auto"}:
            raise ValueError("search_mode must be 'matrix', 'chunked', or 'auto'.")
        self._search_mode = value

    @property
    def auto_matrix_bytes_threshold(self) -> int:
        return int(self._auto_matrix_bytes_threshold)

    @auto_matrix_bytes_threshold.setter
    def auto_matrix_bytes_threshold(self, value: int) -> None:
        value_i = int(value)
        if value_i <= 0:
            raise ValueError("auto_matrix_bytes_threshold must be > 0.")
        self._auto_matrix_bytes_threshold = value_i

    def search(self, xq: torch.Tensor, k: int) -> tuple[torch.Tensor, torch.Tensor]:
        xq = self._validate_input(xq)
        if k <= 0:
            raise ValueError("k must be a positive integer.")
        if self._ntotal == 0:
            dists = torch.full((xq.shape[0], k), float("inf"), dtype=self.dtype, device=self.device)
            labels = torch.full((xq.shape[0], k), -1, dtype=torch.long, device=self.device)
            if self.metric == "ip":
                dists.fill_(float("-inf"))
            return dists, labels

        mode = self._search_mode
        if mode == "auto":
            element_size = torch.empty((), dtype=self.dtype, device="cpu").element_size()
            est_bytes = int(xq.shape[0]) * int(self._ntotal) * int(element_size)
            mode = "chunked" if est_bytes > int(self._auto_matrix_bytes_threshold) else "matrix"

        if mode == "chunked":
            return self._search_chunked(xq, k)

        scores = self._compute_scores(xq, self._xb)
        if self.metric == "l2":
            dists, idx = torch.topk(scores, min(k, scores.shape[1]), largest=False, dim=1)
        else:  # IP
            dists, idx = torch.topk(scores, min(k, scores.shape[1]), largest=True, dim=1)

        labels = self._ids[idx]
        if idx.shape[1] < k:
            pad_cols = k - idx.shape[1]
            pad_dist = torch.full(
                (scores.shape[0], pad_cols),
                float("inf") if self.metric == "l2" else float("-inf"),
                dtype=self.dtype,
                device=self.device,
            )
            pad_labels = torch.full(
                (scores.shape[0], pad_cols),
                -1,
                dtype=torch.long,
                device=self.device,
            )
            dists = torch.cat([dists, pad_dist], dim=1)
            labels = torch.cat([labels, pad_labels], dim=1)
        return dists, labels

    def _search_chunked(self, xq: torch.Tensor, k: int) -> tuple[torch.Tensor, torch.Tensor]:
        nq = xq.shape[0]
        nb = int(self._ntotal)
        largest = self.metric == "ip"
        fill = float("-inf") if largest else float("inf")

        best_scores = torch.full((nq, k), fill, dtype=self.dtype, device=self.device)
        best_idx = torch.full((nq, k), -1, dtype=torch.long, device=self.device)
        if nb == 0:
            labels = torch.full((nq, k), -1, dtype=torch.long, device=self.device)
            if self.metric == "l2":
                best_scores.fill_(float("inf"))
            return best_scores, labels

        q = xq.to(self.dtype).contiguous()
        q2 = (q * q).sum(dim=1) if self.metric == "l2" else None

        element_size = torch.empty((), dtype=self.dtype, device="cpu").element_size()
        budget_bytes = 256 * 1024 * 1024 if self.device.type == "cuda" else 512 * 1024 * 1024
        chunk = max(1, int(budget_bytes // max(1, nq * element_size)))
        chunk = min(chunk, nb)

        for start in range(0, nb, chunk):
            end = min(nb, start + chunk)
            xb_chunk = self._xb[start:end]
            prod = torch.matmul(q, xb_chunk.transpose(0, 1))
            if self.metric == "l2":
                x2 = self._xb_norms[start:end]
                dist = q2.unsqueeze(1) + x2.unsqueeze(0) - (2.0 * prod)
                dist = dist.clamp_min_(0)
                topk = min(k, dist.shape[1])
                cand_scores, cand_j = torch.topk(dist, topk, largest=False, dim=1)
            else:
                topk = min(k, prod.shape[1])
                cand_scores, cand_j = torch.topk(prod, topk, largest=True, dim=1)

            cand_idx = (start + cand_j).to(torch.long)
            if topk < k:
                pad_cols = k - topk
                cand_scores = torch.cat(
                    [cand_scores, torch.full((nq, pad_cols), fill, dtype=self.dtype, device=self.device)],
                    dim=1,
                )
                cand_idx = torch.cat(
                    [cand_idx, torch.full((nq, pad_cols), -1, dtype=torch.long, device=self.device)],
                    dim=1,
                )

            best_scores, best_idx = self._merge_topk(best_scores, best_idx, cand_scores, cand_idx, k, largest=largest)

        labels = self._ids.index_select(0, best_idx.clamp_min(0).reshape(-1)).reshape(best_idx.shape)
        labels = torch.where(best_idx < 0, torch.full_like(labels, -1), labels)
        return best_scores, labels

    def range_search(self, xq: torch.Tensor, radius: float):
        xq = self._validate_input(xq)
        nq = xq.shape[0]
        if self._ntotal == 0:
            lims = torch.zeros(nq + 1, dtype=torch.long, device=self.device)
            return lims, torch.empty(0, dtype=self.dtype, device=self.device), torch.empty(
                0, dtype=torch.long, device=self.device
            )

        scores = self._compute_scores(xq, self._xb)
        if self.metric == "l2":
            mask = scores <= radius
        else:
            mask = scores >= radius

        vals = scores[mask]
        repeated_ids = self._ids.unsqueeze(0).expand(nq, -1)
        labels = repeated_ids[mask]
        counts = mask.sum(dim=1, dtype=torch.long)
        lims = torch.zeros(nq + 1, dtype=torch.long, device=self.device)
        lims[1:] = torch.cumsum(counts, dim=0)
        return lims, vals, labels

    def _merge_topk(
        self,
        best_scores: torch.Tensor,
        best_idx: torch.Tensor,
        cand_scores: torch.Tensor,
        cand_idx: torch.Tensor,
        k: int,
        *,
        largest: bool,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        merged_scores = torch.cat([best_scores, cand_scores], dim=1)
        merged_idx = torch.cat([best_idx, cand_idx], dim=1)
        new_scores, pos = torch.topk(merged_scores, k, largest=largest, dim=1)
        new_idx = torch.gather(merged_idx, 1, pos)
        return new_scores, new_idx

    def reset(self) -> None:
        self._xb = torch.empty((0, self.d), dtype=self.dtype, device=self.device)
        self._xb_norms = torch.empty(0, dtype=self.dtype, device=self.device)
        self._ids = torch.empty(0, dtype=torch.long, device=self.device)
        self._ntotal = 0

    # ------------------------------------------------------------------ #
    # Helpers
    # ------------------------------------------------------------------ #
    def _tensor_attributes(self):
        return ("_xb", "_xb_norms", "_ids")

    def _append(self, xb: torch.Tensor, ids: torch.Tensor) -> None:
        self._xb = torch.cat([self._xb, xb], dim=0)
        if self.metric == "l2":
            norms = (xb * xb).sum(dim=1)
            self._xb_norms = torch.cat([self._xb_norms, norms], dim=0)
        self._ids = torch.cat([self._ids, ids.to(torch.long)], dim=0)
        self._ntotal = self._xb.shape[0]

    def _compute_scores(self, xq: torch.Tensor, xb: torch.Tensor) -> torch.Tensor:
        if xb.shape[0] == 0:
            return torch.empty(xq.shape[0], 0, dtype=self.dtype, device=self.device)
        if self.metric == "l2":
            xq_cast = xq.to(self.dtype)
            xb_cast = xb.to(self.dtype)
            xq_norm = (xq_cast * xq_cast).sum(dim=1, keepdim=True)
            xb_norm = self._xb_norms.unsqueeze(0) if self._xb_norms.numel() else (xb_cast * xb_cast).sum(dim=1).unsqueeze(0)
            dist = xq_norm + xb_norm - (2.0 * (xq_cast @ xb_cast.t()))
            return dist.clamp_min_(0)
        xq_cast = xq.to(self.dtype)
        xb_cast = xb.to(self.dtype)
        return xq_cast @ xb_cast.t()

    def _validate_ids(self, ids: torch.Tensor, expected: int) -> torch.Tensor:
        if not isinstance(ids, torch.Tensor):
            raise TypeError("ids must be a torch.Tensor of dtype long.")
        if ids.ndim != 1 or ids.shape[0] != expected:
            raise ValueError(f"ids must be 1-D with length {expected}.")
        if ids.dtype not in {torch.int64, torch.long}:
            raise ValueError("ids dtype must be torch.long.")
        return ids



class IndexFlatL2(IndexFlat):
    def __init__(
        self,
        d: int,
        *,
        device: torch.device | str | None = None,
        dtype: torch.dtype = torch.float32,
    ) -> None:
        super().__init__(d, metric="l2", device=device, dtype=dtype)


class IndexFlatIP(IndexFlat):
    def __init__(
        self,
        d: int,
        *,
        device: torch.device | str | None = None,
        dtype: torch.dtype = torch.float32,
    ) -> None:
        super().__init__(d, metric="ip", device=device, dtype=dtype)



from __future__ import annotations

import copy
from abc import ABC, abstractmethod
from typing import Iterable, Literal, Sequence

import torch

MetricType = Literal["l2", "ip"]


class IndexBase(ABC):
    """Shared functionality for IVF/Flat indexes."""

    _VALID_METRICS: tuple[MetricType, ...] = ("l2", "ip")

    def __init__(
        self,
        d: int,
        *,
        metric: MetricType = "l2",
        device: torch.device | str | None = None,
        dtype: torch.dtype = torch.float32,
    ) -> None:
        if d <= 0:
            raise ValueError("dimension `d` must be a positive integer.")
        if metric not in self._VALID_METRICS:
            raise ValueError(f"metric must be one of {self._VALID_METRICS}, got {metric!r}")

        self._d = int(d)
        self._metric: MetricType = metric
        self._dtype = dtype
        self._device = self._resolve_device(device)
        self._ntotal = 0
        self._is_trained = False

    # --------------------------------------------------------------------- #
    # Required interface
    # --------------------------------------------------------------------- #
    @abstractmethod
    def train(self, xb: torch.Tensor) -> None:  # pragma: no cover - abstract method
        raise NotImplementedError

    @abstractmethod
    def add(self, xb: torch.Tensor) -> None:  # pragma: no cover - abstract method
        raise NotImplementedError

    @abstractmethod
    def add_with_ids(self, xb: torch.Tensor, ids: torch.Tensor) -> None:  # pragma: no cover
        raise NotImplementedError

    @abstractmethod
    def search(self, xq: torch.Tensor, k: int) -> tuple[torch.Tensor, torch.Tensor]:  # pragma: no cover
        raise NotImplementedError

    @abstractmethod
    def range_search(self, xq: torch.Tensor, radius: float) -> Sequence[torch.Tensor]:  # pragma: no cover
        raise NotImplementedError

    @abstractmethod
    def reset(self) -> None:  # pragma: no cover - abstract method
        raise NotImplementedError

    # --------------------------------------------------------------------- #
    # Public helpers
    # --------------------------------------------------------------------- #
    @property
    def d(self) -> int:
        return self._d

    @property
    def metric(self) -> MetricType:
        return self._metric

    @property
    def dtype(self) -> torch.dtype:
        return self._dtype

    @property
    def device(self) -> torch.device:
        return self._device

    @property
    def ntotal(self) -> int:
        return self._ntotal

    @property
    def is_trained(self) -> bool:
        return self._is_trained

    def to(self, device: torch.device | str | None) -> "IndexBase":
        """Return a copy of the index on the requested device."""
        if device is None:
            return self
        target = self._resolve_device(device)
        cloned = copy.deepcopy(self)
        cloned._apply_to_tensors(lambda tensor: tensor.to(target))
        cloned._device = target
        return cloned

    def cpu(self) -> "IndexBase":
        return self.to(torch.device("cpu"))

    def cuda(self, device: int | str | torch.device | None = None) -> "IndexBase":
        target = torch.device(device) if device is not None else torch.device("cuda")
        return self.to(target)

    def rocm(self, device: int | str | torch.device | None = None) -> "IndexBase":
        if not torch.version.hip:
            raise RuntimeError("Current PyTorch build does not support ROCm.")
        target = torch.device(device) if device is not None else torch.device("cuda")
        return self.to(target)

    def dml(self, device: int | str | torch.device | None = None) -> "IndexBase":
        target = torch.device(device) if device is not None else torch.device("dml")
        return self.to(target)


    # ------------------------------------------------------------------ #
    # Utilities for subclasses
    # ------------------------------------------------------------------ #
    def _resolve_device(self, device: torch.device | str | None) -> torch.device:
        if device is None:
            # Mirror torch default device (currently CPU) without forcing CUDA init.
            return torch.empty(0).device
        return torch.device(device)

    def _tensor_attributes(self) -> Iterable[str]:
        """Override to return tensor attribute names to be moved in `to()`."""
        return ()


    def _apply_to_tensors(self, fn) -> None:
        for name in self._tensor_attributes():
            tensor = getattr(self, name, None)
            if tensor is None:
                continue
            setattr(self, name, fn(tensor))

    def _validate_input(self, x: torch.Tensor) -> torch.Tensor:
        if not isinstance(x, torch.Tensor):
            raise TypeError("Expected torch.Tensor input.")
        if x.ndim != 2:
            raise ValueError(f"Expected 2D tensor, got shape {tuple(x.shape)}.")
        if x.shape[1] != self._d:
            raise ValueError(f"Tensor dimension mismatch: expected {self._d}, got {x.shape[1]}.")
        if x.dtype not in {torch.float16, torch.float32, torch.bfloat16}:
            raise ValueError("Tensor dtype must be float16/float32/bfloat16.")
        return x.to(self._device)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(d={self._d}, metric={self._metric}, "
            f"device='{self._device}', dtype={self._dtype}, ntotal={self._ntotal}, "
            f"is_trained={self._is_trained})"
        )
